import os
import shutil
from pathlib import Path
from typing import List, Optional

# From Third party
import click
from colorama import Fore

# From This Project
from fdn.fdnlib import fdnutils, utils
from fdn.fdnlib.fdncfg import gParamDict as ugPD


@click.command(
    context_settings={
        "ignore_unknown_options": True,
        "help_option_names": ["-h", "--help"],
        "max_content_width": shutil.get_terminal_size()[0]
    })
@click.argument("path",
                required=False,
                type=click.Path(exists=True),
                nargs=-1)
@click.option("--max-depth",
              "-d",
              default=1,
              type=int,
              help=f"Set travel directory tree with max depth.",
              show_default=True)
@click.option("--file-type",
              "-t",
              default="file",
              type=click.Choice(["file", "dir"]),
              help=f"Set type.If 'file',operations are only valid for file,If 'dir',"
                   f"operations are only valid for directory.",
              show_default=True)
@click.option("--in-place",
              "-i",
              default=False,
              type=bool,
              is_flag=True,
              help=f"Changes file name in place.",
              show_default=True)
@click.option("--confirm",
              "-c",
              default=False,
              type=bool,
              is_flag=True,
              help=f"Need confirmation before change to take effect.",
              show_default=True)
@click.option("--is-link",
              "-l",
              default=False,
              type=bool,
              is_flag=True,
              help=f"Follow the real path of a link.",
              show_default=True)
@click.option("--full-path",
              "-f",
              default=False,
              type=bool,
              is_flag=True,
              help=f"Show full path of file.Relative to the input path.",
              show_default=True)
@click.option("--absolute-path",
              "-a",
              default=False,
              type=bool,
              is_flag=True,
              help=f"Show absolute path of file.",
              show_default=True)
@click.option("--roll-back",
              "-r",
              default=False,
              type=bool,
              is_flag=True,
              help=f"To roll back changed file names.",
              show_default=True)
@click.option("--overwrite",
              "-o",
              default=False,
              type=bool,
              is_flag=True,
              help=f"Overwrite exist files.",
              show_default=True)
@click.option("--pretty",
              "-p",
              default=False,
              type=bool,
              is_flag=True,
              help=f"Try to pretty output.",
              show_default=True)
@click.option("--enhanced-display",
              "-e",
              default=False,
              type=bool,
              is_flag=True,
              help=f"Enhanced display output.",
              show_default=True)
@click.option("--verbose",
              default=False,
              type=bool,
              is_flag=True,
              help=f"Display more information.",
              show_default=True,
              hidden=True)
@click.option("--debug",
              default=False,
              type=bool,
              is_flag=True,
              hidden=True)
@click.option("--external_config",
              "-xf",
              default=False,
              required=False,
              type=bool,
              is_flag=True,
              help=f"Enable external config")
@click.version_option(version="2022.5.5.3240")
def ufn(path: Optional[List[Path]], max_depth: int, file_type: str,
        in_place: bool, confirm: bool, is_link: bool, full_path: bool,
        absolute_path: bool, roll_back: bool, overwrite: bool, pretty: bool,
        enhanced_display: bool, verbose: bool, debug: bool, external_config: bool):
    """Files in PATH will be changed file names unified.
    """
    if verbose:
        print(f"verbose mode")
    if debug:
        print(f"debug mode")
    if not path:
        ugPD["path"] = ["."]
    else:
        ugPD["path"] = path
    ugPD["max_depth"] = max_depth
    ugPD["type"] = file_type
    ugPD["in_place_flag"] = in_place
    ugPD["need_confirmation_flag"] = confirm
    ugPD["is_link"] = is_link
    ugPD["full_path_flag"] = full_path
    ugPD["absolute_path_flag"] = absolute_path
    ugPD["overwrite_flag"] = overwrite
    ugPD["pretty_flag"] = pretty
    ugPD["enhanced_display_flag"] = enhanced_display
    ugPD["all_in_place_flag"] = False
    ugPD["latest_confirm"] = utils.unify_confirm(
    )  # Parameter is Null to get default return
    ugPD["target_appeared"] = False

    if external_config:
        fdnutils.load_config()

    for pth in ugPD["path"]:
        pth = (pth[:-1] if pth.endswith(os.sep) else pth)
        if os.path.isfile(pth) and fdnutils.type_matched(pth):
            if not roll_back:
                fdnutils.one_file_ufn(pth)
            else:
                fdnutils.one_file_rbk(pth)
        elif os.path.isdir(pth):
            if not roll_back:
                fdnutils.one_dir_ufn(pth)
            else:
                fdnutils.one_dir_rbk(pth)
        else:
            click.echo(f"{Fore.RED}Not valid:{pth}{Fore.RESET}")

    if (not ugPD["in_place_flag"]) and (
            not ugPD["need_confirmation_flag"]) and (ugPD["target_appeared"]):
        cols, _ = shutil.get_terminal_size(fallback=(79, 23))
        click.echo("*" * cols)
        click.echo("In order to take effect,add option '-i' or '-c'")
        ugPD["target_appeared"] = False

# TODO: terminology not a word but surround different type as a word
# TODO: global version how to
# TODO: **support edit config data
# TODO: display config data
# TODO: display total summary
# TODO: support regular expression input path or directory as path argument
# TODO: exclude spec directory or file type
# TODO: %,'s,How to ...

# TODO: unify full_path request to a function ...
# TODO: Embed nltk ,but need download first time ,why?
# TODO: multi dir may occur same name conflict so record path ID?
# TODO: support user select roll back when multi target

# TODO: bug:mac osx begin with ._ hidden check wrong
# TODO: bug:fd -x can not get terminal size .so fall back and no color display
# TODO: autocomplete
# TODO: launch slowly



import base64
from typing import Optional, Union

from cryptography.fernet import Fernet
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC


class Crypt:
    def __init__(self, salt: str = 'HobbyMarks'):
        self.salt = salt.encode('UTF-8')
        self.kdf = PBKDF2HMAC(algorithm=hashes.SHA256(),
                              length=32,
                              salt=self.salt,
                              iterations=100000)

    def get_key(self, passwd: str) -> bytes:
        return base64.urlsafe_b64encode(self.kdf.derive(
            passwd.encode("UTF-8")))

    def encrypt(self, plaintext: Union[str, bytes],
                passwd: str) -> Optional[bytes]:
        key = self.get_key(passwd)
        if type(plaintext) is str:
            return Fernet(key).encrypt(plaintext.encode("UTF-8"))
        elif type(plaintext) is bytes:
            return Fernet(key).encrypt(plaintext)
        else:
            return None

    def decrypt(self, ciphertext: Union[bytes, str],
                passwd: str) -> Optional[bytes]:
        key = self.get_key(passwd)
        if type(ciphertext) is bytes:
            return Fernet(key).decrypt(ciphertext)
        elif type(ciphertext) is str:
            return Fernet(key).decrypt(ciphertext.encode("UTF-8"))
        else:
            return None


def encrypt_b64_str(plaintext: str, passwd: str, salt: str = "") -> str:
    if salt == "":
        crypt_ins = Crypt()
    else:
        crypt_ins = Crypt(salt)
    enc_bytes = crypt_ins.encrypt(plaintext, passwd)
    return base64.b64encode(enc_bytes).decode("UTF-8")


def b64_str_decrypt(ciphertext: str, passwd: str, salt: str = "") -> str:
    if salt == "":
        crypt_ins = Crypt()
    else:
        crypt_ins = Crypt(salt)
    enc_bytes = base64.b64decode(ciphertext)
    return crypt_ins.decrypt(enc_bytes, passwd).decode("UTF-8")



import importlib.resources
import json
import os
import shutil
from pathlib import Path

# From Third party
import nltk
from nltk.corpus import words

# From Project
import fdn.data

gParamDict: dict = {}
with importlib.resources.path("fdn.data", "config.json") as cfg_path:
    with open(cfg_path, encoding="UTF-8") as fh:
        gParamDict.update(json.load(fh))

# TODO: can optimize to not require nltk package
nltk_path = os.path.dirname(fdn.data.__file__)
if os.path.isdir(nltk_path):
    nltk.data.path.append(nltk_path)
    if not os.path.isdir(
            words_path := os.path.join(nltk_path, "corpora", "words")) or len(
        os.listdir(words_path)) == 0:
        if os.path.isfile(
                zip_path := os.path.join(nltk_path, "corpora", "words.zip")):
            shutil.unpack_archive(
                zip_path,
                os.path.join(nltk_path, "corpora"))
        else:
            nltk.download("words", download_dir=nltk_path)
else:
    try:
        from nltk.corpus import words

        gParamDict["LowerCaseWordSet"] = set(
            list(map(lambda x: x.lower(), words.words())))
    except LookupError:
        nltk.download("words")
gParamDict["LowerCaseWordSet"] = set(
    list(map(lambda x: x.lower(), words.words())))

gParamDict["record_path"] = os.path.join(Path.home(), ".fdn")
Path(gParamDict["record_path"]).mkdir(parents=True, exist_ok=True)
gParamDict["db_path"] = os.path.join(gParamDict["record_path"], "rdsa.db")
gParamDict["config_path"] = os.path.join(gParamDict["record_path"], "config.json")
if not os.path.isfile(gParamDict["config_path"]):
    _configDict = {"ReplacedDictionary": {}, "RemainedList": []}
    json.dumps(_configDict)
    with open(file=gParamDict["config_path"], mode="w", encoding="UTF-8") as fh:
        json.dump(_configDict, fh)



"""fdn use to uniformly change file or directory names and also support
rollback these operations. """
from __future__ import print_function

import sys

# From Third party
import click
import colorama
from colorama import Back, Fore, Style

# From Project
from fdn.fdnlib.fdncli import ufn

__version__ = "2022.5.5.3240"


def main() -> None:
    try:
        colorama.init()
        if (sys.version_info.major, sys.version_info.minor) < (3, 8):
            click.echo(
                f"{Fore.RED}current is {sys.version},\n"
                f"{Back.WHITE}Please upgrade to >=3.8.{Style.RESET_ALL}")
            sys.exit()
        #######################################################################
        ufn()
        #######################################################################
    finally:
        colorama.deinit()


def run_main():
    try:
        sys.exit(main())
    except Exception as e:
        sys.stderr.write(f"Error:{str(e)}\n")
        sys.exit(1)


if __name__ == '__main__':
    run_main()



import difflib
import hashlib
import os
from pathlib import Path
from typing import Generator, List, Optional, Tuple, Union

# From Third Party
import click
from colorama import Fore
from wcwidth import wcswidth


def is_hidden(f_path: Path) -> bool:
    """
    Check file is hidden or not
    :param f_path: string,file path
    :return: True if is hidden file,False if is not hidden file
    """
    if os.name == "nt":
        import win32api
        import win32con
    if os.name == "nt":
        attribute = win32api.GetFileAttributes(str(f_path))
        return attribute & (win32con.FILE_ATTRIBUTE_HIDDEN
                            | win32con.FILE_ATTRIBUTE_SYSTEM)
    else:
        return os.path.basename(f_path).startswith(".")  # linux, osx


def depth_walk(
        top_path: Path,
        top_down: bool = False,
        follow_links: bool = False,
        max_depth: int = 1
) -> Optional[Generator[Tuple[str], List[str], List[str]]]:
    if str(max_depth).isnumeric():
        max_depth = int(max_depth)
    else:
        max_depth = 1
    try:
        names = os.listdir(top_path)
    except FileNotFoundError:
        click.echo(f"Warning:{top_path} not found.")
        return None
    except PermissionError:
        click.echo(f"Warning:{top_path} no permissions.")
        return None
    dirs, non_dirs = [], []
    for name in names:
        if os.path.isdir(os.path.join(top_path, name)):
            dirs.append(name)
        else:
            non_dirs.append(name)
    if top_down:
        yield top_path, dirs, non_dirs
    if max_depth > 1:
        for name in dirs:
            new_path = Path(os.path.join(top_path, name))
            if follow_links or not os.path.islink(new_path):
                for x in depth_walk(new_path, top_down, follow_links,
                                    1 if max_depth == 1 else max_depth - 1):
                    yield x
    if not top_down:
        yield top_path, dirs, non_dirs


def rich_style(
        original: str,
        processed: str,
        pretty: bool = False,
        enhanced_display: bool = False
) -> Union[Tuple[None, None], Tuple[str, str]]:
    if (type(original) is not str) or (type(processed) is not str):
        return None, None

    def _f_d(s: str = "", f_d=None):
        return f_d + s + Fore.RESET

    def _c_f(s: str = "") -> str:
        if pretty:
            return s
        else:
            return ""

    a = original
    b = processed
    a_list = []
    b_list = []
    c_f = ' '
    e = enhanced_display
    for tag, i1, i2, j1, j2 in difflib.SequenceMatcher(None, a,
                                                       b).get_opcodes():
        if tag == "delete":
            a_list.append(_f_d(a[i1:i2].replace(" ", "▯"), Fore.RED))
            b_list.append(_c_f(c_f * wcswidth(a[i1:i2])))
        elif tag == "equal":
            a_list.append(_f_d(a[i1:i2], Fore.BLACK) if e else a[i1:i2])
            b_list.append(_f_d(b[j1:j2], Fore.BLACK) if e else b[j1:j2])
        elif tag == "replace":
            a_w = wcswidth(a[i1:i2])
            b_w = wcswidth(b[j1:j2])
            if a_w > b_w:
                a_list.append(_f_d(a[i1:i2].replace(" ", "▯"), Fore.RED))
                b_list.append(
                    _c_f(c_f * (a_w - b_w)) +
                    _f_d(b[j1:j2].replace(" ", "▯"), Fore.GREEN))

            elif a_w < b_w:
                a_list.append(
                    _c_f(c_f * (b_w - a_w)) +
                    _f_d(a[i1:i2].replace(" ", "▯"), Fore.RED))

                b_list.append(_f_d(b[j1:j2].replace(" ", "▯"), Fore.GREEN))
            else:
                a_list.append(_f_d(a[i1:i2].replace(" ", "▯"), Fore.RED))
                b_list.append(_f_d(b[j1:j2].replace(" ", "▯"), Fore.GREEN))
        elif tag == "insert":
            a_list.append(_c_f(c_f * wcswidth(b[j1:j2])))
            b_list.append(_f_d(b[j1:j2].replace(" ", "▯"), Fore.GREEN))
    return "".join(a_list), "".join(b_list)


def unify_confirm(x: str = "") -> str:
    return {
        "y": "yes",
        "yes": "yes",
        "n": "no",
        "no": "no",
        "A": "all",
        "all": "all",
        "q": "quit",
        "quit": "quit"
    }.get(x, "no")


def path_exist(path: str) -> bool:
    bn = os.path.basename(path)
    bns = os.listdir(os.path.dirname(path))
    if bn in bns:
        return True
    else:
        return False


###############################################################################


def sha2_id(s: str) -> str:
    return hashlib.sha256(s.encode("UTF-8")).hexdigest()



import json
import os
import re
import string
import sys
from pathlib import Path
from typing import List, Optional, Tuple, Union

# From Third Party
import click
import emoji
from colorama import Back
from unidecode import unidecode

# From This Project
from fdn.fdnlib import ucrypt, utils
from fdn.fdnlib.fdncfg import gParamDict as ugPD
from fdn.fdnlib.udb import FDNDB


def _replace_char(s: str) -> str:
    """
    Replace char in s ,when the char in "BeReplacedCharDictionary"
    :param s:string,if "BeReplacedChar" in s ,then will be replaced
    :return:string,be replaced
    """

    def _mask_ruw(_s: str) -> Tuple[List[str], List[bool]]:
        """
        Mask a input string by RemainUnchangedWordList value
        :param _s: string,input string to be masked
        :return: word list and mask list
        """
        ruw_list = ugPD["RemainUnchangedWordList"][:]
        ruw_list.extend([*ugPD["TerminologyDictionary"]])
        re_str = "|".join([re.escape(ruw) for ruw in ruw_list])
        w_list = re.split(f"({re_str})", _s)
        m_list = []
        for elm in w_list:
            if elm in ruw_list:
                m_list.append(True)
            else:
                m_list.append(False)
        return w_list, m_list

    c_dict = ugPD["BeReplacedCharDictionary"]
    sep_c = ugPD["SeparatorChar"]
    re_cns = re.compile(
        f"[{sep_c}]+")  # To recognize continuous separator char
    word_list, mask_list = _mask_ruw(s)

    new_word_list = []
    for word, mask in zip(word_list, mask_list):
        if mask:
            new_word_list.append(word)
        else:
            # TODO: May Be Not Highest Efficiency
            for c_key, c_value in c_dict.items():
                if c_key in word:  # If need replace ,then go ...
                    word = word.replace(c_key, c_value)
            tcs = []
            for c in word:
                if c in emoji.UNICODE_EMOJI_ENGLISH:
                    tcs.append(sep_c)
                else:
                    tcs.append(c)
            word = ''.join(tcs)
            word = re_cns.sub(sep_c, word)  # When continuous separator,remain
            # only one
            new_word_list.append(word)
    return "".join(new_word_list)


def _del_head_tail_sep(s: str) -> str:
    """
    Delete the beginning and the end separator in string.
    :param s: string
    :return: string
    """
    sep_char = ugPD["SeparatorChar"]
    if s.startswith(sep_char):
        s = sep_char.join(s.split(sep_char)[1:])
    if s.endswith(sep_char):
        s = sep_char.join(s.split(sep_char)[0:-1])
    return s


def _process_head_tail(s: str) -> str:
    # Capitalize The First Letter
    # if s[0].islower():
    #     return s[0].upper() + s[1:]
    return s


def _del_white_space(s: str) -> str:
    """
    All white space will be replaced by Separator Char.
    :param s:string,input string
    :return:string,all whitespace in input string will be replaced by
    Separator Char then return the processed input string
    """
    sep_char = ugPD["SeparatorChar"]
    return sep_char.join(s.split())


def _process_terminology(s: str) -> str:
    """

    :param s:
    :return:
    """

    def _swt(_s: str) -> Optional[str]:
        """
        Check if the s startswith a terminology word.If the s startswith a
        terminology word then return the word lower case ,else return None.
        :param _s:string,input
        :return:None or string
        """
        t_d = ugPD["TerminologyDictionary"]
        for k in t_d.keys():
            if _s.lower().startswith(k):
                return k
        return None

    sep_char = ugPD["SeparatorChar"]
    word_list = s.split(sep_char)
    term_dict = ugPD["TerminologyDictionary"]
    new_word_list = []
    for word in word_list:
        if word.lower() in term_dict.keys():
            new_word_list.append(term_dict[word.lower()])
        elif key := _swt(word):
            if word.lower() in ugPD["LowerCaseWordSet"]:
                new_word_list.append(word)
            else:
                new_word = term_dict[key] + word[len(key):]
                new_word_list.append(new_word)
        else:
            new_word_list.append(word)
    return sep_char.join(new_word_list)


def _asc_head(s: str) -> str:
    """
    If s not starts with ascii letters ,will create at most "ASCLen" length
    asc letters then return ,else return null string.
    :param s:string
    :return:string ,null string or at most "ASCLen" length ascii letters
    """
    lmt_len = ugPD["ASCLen"]
    sep_char = ugPD["SeparatorChar"]
    head_chars = ugPD["HeadChars"]
    if s[0] in head_chars:
        return ""
    word = s.split(sep_char)[0]
    if len(word) > lmt_len:
        word = word[0:lmt_len]
    new_word = ""
    for c in word:
        if c not in head_chars:
            new_word += c
        else:
            break
    return "".join([elm[0].upper() for elm in unidecode(new_word).split()])


def _process_word(s: str) -> str:
    sep_char = ugPD["SeparatorChar"]
    word_list = s.split(sep_char)
    new_word_list = []
    word_set = ugPD["LowerCaseWordSet"]
    for word in word_list:
        if word.lower() in word_set:
            new_word_list.append(string.capwords(word))
        else:
            new_word_list.append(word)
    return sep_char.join(new_word_list)


def _confirm(p_i: str = "") -> str:
    v = click.prompt(f"{p_i}\nPlease confirm(y/n/A/q)",
                     type=click.Choice(
                         ["y", "yes", "n", "no", "A", "all", "q", "quit"]),
                     show_choices=False,
                     default=ugPD["latest_confirm"])

    return utils.unify_confirm(v)


def _in_place(p_i: str = "") -> bool:
    if ugPD["all_in_place_flag"]:
        return True
    if ugPD["need_confirmation_flag"]:
        if (c := _confirm(p_i)) == utils.unify_confirm("yes"):
            ugPD["latest_confirm"] = c
            return True
        elif c == utils.unify_confirm("no"):
            ugPD["latest_confirm"] = c
            return False
        elif c == utils.unify_confirm("all"):
            ugPD["latest_confirm"] = c
            ugPD["all_in_place_flag"] = True
            return True
        elif c == utils.unify_confirm("quit"):
            ugPD["latest_confirm"] = c
            sys.exit()  # TODO: roughly process ...
    else:
        if ugPD["in_place_flag"]:
            return True
        else:
            return False


def _out_info(file: str, new_name: str, take_effect: bool = False) -> None:
    def _b_d(s="", b_d=""):
        if ugPD["enhanced_display_flag"]:
            return b_d + s + Back.RESET
        else:
            return s

    rich_org, rich_proc = utils.rich_style(
        file,
        new_name,
        pretty=ugPD["pretty_flag"],
        enhanced_display=ugPD["enhanced_display_flag"])
    if ugPD["AlternateFlag"]:
        click.echo(" " * 3 + (_b_d(rich_org, Back.WHITE)))
    else:
        click.echo(" " * 3 + (_b_d(rich_org, Back.LIGHTWHITE_EX)))
    if take_effect:
        if ugPD["AlternateFlag"]:
            click.echo("==>" + (_b_d(rich_proc, Back.WHITE)))
        else:
            click.echo("==>" + (_b_d(rich_proc, Back.LIGHTWHITE_EX)))
    else:
        if ugPD["AlternateFlag"]:
            click.echo("-->" + (_b_d(rich_proc, Back.WHITE)))
        else:
            click.echo("-->" + (_b_d(rich_proc, Back.LIGHTWHITE_EX)))
    ugPD["AlternateFlag"] = not ugPD["AlternateFlag"]


def type_matched(f_path: str) -> bool:
    if (ugPD["type"] == "file") and (os.path.isfile(f_path)):
        return True
    if (ugPD["type"] == "dir") and (os.path.isdir(f_path)):
        return True
    return False


def one_file_ufn(f_path: Path) -> None:
    if os.path.islink(f_path) != ugPD["is_link"]:
        return None

    subdir, file = os.path.split(f_path)
    root, ext = os.path.splitext(file)
    # all whitespace replace by sep_char
    root = _del_white_space(root)
    # replace characters by defined Dictionary
    root = _replace_char(root)
    # Capwords only when word in wordsSet
    root = _process_word(root)
    # process Head and Tail
    root = _process_head_tail(root)
    root = _del_head_tail_sep(root)  # special terminology may conflict
    # Pretty Terminology
    root = _process_terminology(root)
    # Add ascii head at the beginning
    new_name = _asc_head(root) + root + ext

    # Create full path
    new_path = os.path.join(subdir, new_name)

    if new_name == file:
        return None

    _ap = ugPD["absolute_path_flag"]
    _fp = ugPD["full_path_flag"]

    if _ap:
        ip = _in_place(os.path.abspath(f_path))
    elif _fp:
        ip = _in_place(str(f_path))
    else:
        ip = _in_place(file)
    if ip:
        if utils.path_exist(os.path.abspath(new_path)) and (not ugPD["overwrite_flag"]):
            click.echo(f"{Back.RED}Exist:{Back.RESET}"
                       f"{new_path if _fp else new_name}\n"
                       f"Skipped:{f_path if _fp else file}\n"
                       f"With option '-o' to enable overwrite.")
            return None
        else:
            log_to_db(cur_name=file, new_path=Path(new_path))
            os.rename(f_path, new_path)
    if _ap:
        _out_info(str(os.path.abspath(f_path)),
                  os.path.abspath(new_path),
                  take_effect=ip)
    elif _fp:
        _out_info(str(f_path), new_path, take_effect=ip)
    else:
        _out_info(file, new_name, take_effect=ip)

    ugPD["target_appeared"] = True


def one_dir_ufn(tgt_path: Path) -> None:
    for subdir, dirs, files in utils.depth_walk(top_path=tgt_path,
                                                max_depth=ugPD["max_depth"]):
        if ugPD["type"] == "file":
            for file in files:
                f_path = Path(os.path.join(subdir, file))
                if utils.is_hidden(f_path):
                    continue
                if os.path.isfile(f_path):
                    one_file_ufn(f_path)
        elif ugPD["type"] == "dir":
            for d in dirs:
                f_path = Path(os.path.join(subdir, d))
                if utils.is_hidden(f_path):
                    continue
                if os.path.isdir(f_path):
                    one_file_ufn(f_path)
    if ugPD["type"] == "dir":
        one_file_ufn(tgt_path)


def one_file_rbk(f_path: Path) -> None:
    if os.path.islink(f_path) != ugPD["is_link"]:
        return None

    subdir, file = os.path.split(f_path)
    if not (new_name := _used_name_lookup(file)):
        return None

    # Create full path
    new_path = os.path.join(subdir, new_name)

    if new_name == file:
        return None

    _ap = ugPD["absolute_path_flag"]
    _fp = ugPD["full_path_flag"]

    if _ap:
        ip = _in_place(os.path.abspath(f_path))
    elif _fp := ugPD["full_path_flag"]:
        ip = _in_place(str(f_path))
    else:
        ip = _in_place(file)
    if ip:
        if utils.path_exist(new_path) and (not ugPD["overwrite_flag"]):
            click.echo(f"{Back.RED}Exist:{Back.RESET}"
                       f"{new_path if _fp else new_name}\n"
                       f"Skipped:{f_path if _fp else file}\n"
                       f"With option '-o' to enable overwrite.")
            return None
        else:
            os.rename(f_path, new_path)
    if _ap:
        _out_info(str(os.path.abspath(f_path)),
                  os.path.abspath(new_path),
                  take_effect=ip)
    elif _fp:
        _out_info(str(f_path), new_path, take_effect=ip)
    else:
        _out_info(file, new_name, take_effect=ip)

    ugPD["target_appeared"] = True


def one_dir_rbk(tgt_path: Path) -> None:
    for subdir, dirs, files in utils.depth_walk(top_path=tgt_path,
                                                max_depth=ugPD["max_depth"]):
        if ugPD["type"] == "file":
            for file in files:
                f_path = Path(os.path.join(subdir, file))
                if utils.is_hidden(f_path):
                    continue
                one_file_rbk(f_path)
        elif ugPD["type"] == "dir":
            for d in dirs:
                f_path = Path(os.path.join(subdir, d))
                if utils.is_hidden(f_path):
                    continue
                one_file_rbk(f_path)
    if ugPD["type"] == "dir":
        one_file_rbk(tgt_path)


###############################################################################
def _used_name_lookup(cur_name: str,
                      db_path: Optional[Path] = None,
                      latest: bool = True) -> Union[None, str, List[str]]:
    if not db_path:
        db_path = ugPD["db_path"]
    _cur_id = utils.sha2_id(cur_name)
    db = FDNDB(db_path)
    df = db.checkout_rd(_cur_id)
    if "curCrypt" not in df.columns:
        return None
    rlt = list(df["curCrypt"])
    if len(rlt) == 0:
        return None
    if latest:
        return ucrypt.b64_str_decrypt(rlt[0], cur_name)
    else:
        return [
            ucrypt.b64_str_decrypt(elm, cur_name)
            for elm in list(dict.fromkeys(rlt))
        ]


def log_to_db(cur_name: str,
              new_path: Path,
              db_path: Optional[Path] = None) -> None:
    if not db_path:
        db_path = ugPD["db_path"]
    _f_p = os.path.abspath(new_path)
    _new = os.path.basename(_f_p)
    abs_dp = os.path.dirname(_f_p)
    _cur_id = utils.sha2_id(cur_name)
    _new_id = utils.sha2_id(_new)
    _cur_crypt = ucrypt.encrypt_b64_str(cur_name, _new)
    _sep_dp_id = "_|_".join([utils.sha2_id(d) for d in abs_dp.split(os.sep)])
    _abs_dp_id = utils.sha2_id(abs_dp)
    db = FDNDB(db_path)
    db.insert_rd(_new_id, _cur_id, _cur_crypt, _sep_dp_id, _abs_dp_id)


def load_config() -> None:
    try:
        with open(ugPD["config_path"], encoding="UTF-8") as fh:
            _cfg = json.load(fh)
            for k, v in _cfg["ReplacedDictionary"].items():
                ugPD["BeReplacedCharDictionary"][k] = v
    except Exception as e:
        pass
